"""
Deep Reinforcement learning.
"""

# import
# import pickle
import logging
import numpy as np
import torch
import torch.nn.functional as F
import torch.optim as optim
from Q_MLP import Q_MLP
# import math
import matplotlib.pyplot as plt


def action_filter(state):
    """
    output valid actions for the state
    """
    actions = []
    for i in range(1, len(state)):
        if state[i] == 0:
            actions.append(i)
    if len(actions) == 0:
        actions.append(0)
    return actions

def ess_action_filter(state):
    """_summary_

    Args:
        state (_type_): represents current MG state:
                        [time, sky_condition, SOC]
        F (_type_): Max consecutive hours of discharge possible

    Returns:
        _type_: list of potential actions represented as values
                0=discharge
                1=charge
                2=idle
    """
    #                        D   C   I
    #                        0   1   2  Max hours
    # ESS directives values [1, -.c, 0, F]
    # look at the current SOC for the ESS and decide potential states
    state = list(state)
    SOC = state[2]
    
    Fx = state[-1]
    
    state = tuple(state)
    # print("SOC: ", SOC)
    # if not fully discharged or fully charged anything goes
    # if 0 < SOC < F: 
    if SOC < Fx and SOC > 0:
        # print("all 3")
        return [0, 1, 2]
    # if fully discharged we can only charge(1) or remain idle (2)
    elif SOC >= Fx:
        # print("charge only")
        # return [1, 2]
        return [1]
    # if fully charged we can only discharge(0) or remain idle (0)
    elif SOC == 0:
        # print("discharge only")
        # return [0, 2]
        return [0]
    
def abs_nrm_ess_evaluator(val, mn, mx, **kwargs):
    R = mn
    # normalize the score
    valnrm = R*((val-mn)/(mx-mn))
    return R - abs(valnrm)

def base_ess_reward(val, **kwargs):
    return val

def base_rpf_scale_ess_reward(val, rpf_scl = .2, goal=200, **kwargs):
    if val < 0:
        return val
    elif val > 0:
        return val * rpf_scl
    elif val == 0:
        return goal


def ess_state_action_reward(state, action, ess_values, PV, D, evaluator=max, 
                            verbose=False, **kwargs):
    ess_out = ess_values[action]*ess_values[-1]
    t = state[0]
    k = state[1]
    pv_out, demand = PV[t, k], D[t]
    return evaluator(PV+ess_out - demand, **kwargs)

class Agent:
    """
    Agent used in the DRL class.
    - name: str, name of the agent;
    - actions: a list of all actions;
    - input_size: int, size of the input, len(state);
    - hidden_layers: list, the number of neurons for the hidden layer,
    - output_size: int, size of the output, len(all actions);
    Keyword Arguments:
    - action_filter: function that filters actions for a given state;
    - Q_path: path to learned parameters of pytorch network.
    """

    def __init__(
        self, name, actions, input_size, hidden_layers, output_size,
        learning_rate=1e-3, learn_epoch=1, epsilon=.5, **kwargs
    ):
        super().__init__()
        # name & actions
        self.name, self.actions = name, actions
        # epsilon
        self.epsilon = epsilon
        # ------------- whether to use action_filter ----------------
        if 'action_filter' in kwargs:
            # print("Filtering actions!!")
            self.filter_action = True
            self.action_filter = kwargs['action_filter']
        else:        
            self.filter_action = False
        # ------------------- GPU ----------------
        if torch.cuda.is_available():
            self.dev = "cuda:0"
        else:
            self.dev = "cpu"
        # ----------------- construct network --------------------
        self.hidden_layers = hidden_layers
        self.input_size, self.output_size = input_size, output_size
        # prediction network
        self.Q = Q_MLP(
            hidden_layer_shape=self.hidden_layers,
            input_size=input_size,
            output_size=output_size,
            seed=1
        )
        self.Q.to(self.dev)
        # target network
        self.Q_target = Q_MLP(
            hidden_layer_shape=self.hidden_layers,
            input_size=input_size,
            output_size=output_size,
            seed=1
        )
        self.Q_target.to(self.dev)
        for p in self.Q.parameters():
            p.data.fill_(0)
        for p in self.Q_target.parameters():
            p.data.fill_(0)
        # load previous parameters
        if 'Q_path' in kwargs:
            self.Q.load_state_dict(torch.load(kwargs['Q_path']))
            self.Q.eval()
        # optimizer
        self.optimizer = optim.Adam(self.Q.parameters(), lr=learning_rate)
        # training step
        self.train_step = 1
        self.loss_memory = []
        self.G_memory = []
        self.learn_epoch = learn_epoch

    def load_parameter(self, actor_path, critic_path):
        """
        Load network parameter from path
        """
        self.actor.load_state_dict(torch.load(actor_path))
        self.critic.load_state_dict(torch.load(critic_path))
        return

    def generate_policy(self, states,):
        policy = {}
        for state in states:
            ky = "-".join(  [str(s)  for s in list(state)[:-1] ]   )
            policy[ ky ] = self.take_action(state)
        return policy


    def take_action(self, state, **kwargs):
        """
        take action (make prediction), based on the input state.
        """
        # make state to tensor
        input_seq = torch.tensor(
            state, dtype=torch.float, device=self.dev
        )
        # make a prediction
        self.Q.eval()
        with torch.no_grad():
            output_seq = list(self.Q(input_seq))
        self.Q.train()
        # valid_actions
        if self.filter_action:
            valid_actions = self.action_filter(state)
            # print("State: {}, Actions: {}".format(state, valid_actions))
        else:
            valid_actions = self.actions
        # get a random number between 0, 1
        if np.random.random() > self.epsilon:
            return valid_actions[np.argmax([
                output_seq[self.actions.index(i)]
                for i in valid_actions
            ])]
        else:
            return valid_actions[np.random.choice(
                range(len(valid_actions)), size=1, replace=False,
                p=[1 / len(valid_actions)] * len(valid_actions)
            )[0]]

    def simulate_action(self, state):
        """
        take action (for simulation), based on the input state.
        """
        # make state to tensor
        input_seq = torch.tensor(
            state, dtype=torch.float, device=self.dev
        )
        # make a prediction
        self.Q.eval()
        with torch.no_grad():
            output_seq = list(self.Q(input_seq))
        self.Q.train()
        # valid_actions
        if self.filter_action:
            valid_actions = self.action_filter(state)
        else:
            valid_actions = self.actions
        return valid_actions[np.argmax([
            output_seq[self.actions.index(i)]
            for i in valid_actions
        ])]

    def learn(self, memory, discount_factor):
        """
        train the network
        """
        # memories
        state_memory = memory[0]
        new_state_memory = memory[1]
        delta_memory = memory[2]
        action_memory = memory[3]
        reward_memory = memory[4]
        # action index
        action_ind = torch.tensor([
            [self.actions.index(a)] for a in action_memory
        ], device=self.dev)  # .flatten()
        # while True:
        for train_iter in range(self.learn_epoch):
            # set zero grad
            self.optimizer.zero_grad()
            # make a prediction
            Q_pred = self.Q(
                torch.FloatTensor(state_memory).to(self.dev)
            ).gather(1, action_ind).flatten()
            # calculate the target
            Q_targ_future = self.Q_target(
                torch.FloatTensor(new_state_memory).to(self.dev)
            ).detach().max(1)[0]
            # target
            Q_targ = torch.FloatTensor([
                reward_memory[i] if delta_memory[i]
                else reward_memory[i] + discount_factor * Q_targ_future[i]
                for i in range(len(reward_memory))
            ])
            Q_pred = Q_pred.to(self.dev)
            Q_targ = Q_targ.to(self.dev)
            # loss
            loss = F.mse_loss(Q_pred, Q_targ)
            # backpropogate
            loss.backward()
            self.optimizer.step()
        self.loss_memory.append(loss.to('cpu').detach().numpy())
        self.train_step += 1
        # soft update the target network
        if self.train_step % 1 == 0:
            self.__soft_update(self.Q, self.Q_target, 0.001)
        return

    def __soft_update(self, Q, Q_target, tau):
        """
        Soft update model parameters:
            θ_target = τ*θ_trained + (1 - τ)*θ_target;
        Q: weights will be copied from;
        Q_target: weights will be copied to;
        tau: interpolation parameter.
        """
        for q_target, q in zip(Q_target.parameters(), Q.parameters()):
            q_target.data.copy_(
                tau * q.data + (1.0 - tau) * q_target.data
            )
        return
    
    def plot_G(self, window=20, start_ind=0, sample=1, dir='', title="Average Return", 
                  TFD={}, close_me=False):
        """
        plot return using time window
        """
        G_plot = {}
        ind = window
        print("size of gmem: ", len(self.G_memory))
        while ind <= len(self.G_memory):
            # sample
            if ind % sample == 0:
                G_plot[ind - 1] = np.mean([
                    self.G_memory[i] for i in range(ind - window, ind, 1)
                ])
            ind += 1
        # plot
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(8, 6))
        ax.plot(
            list(G_plot.keys())[start_ind:],
            list(G_plot.values())[start_ind:],
            'b-'
        )
        ax.set_xlabel('Epochs')
        ax.set_ylabel('G')
        ax.set_title(title, fontdict=TFD)
        fig.tight_layout()
        fig.savefig('{}{}_G.png'.format(dir, self.name), dpi=600)
        if close_me:
            plt.close()
        return
    
    def collect_loss(self, window, sample, ):
        loss_plot = {}
        ind = window
        while ind <= len(self.loss_memory):
            # when we hit a sample size 
            # get the average over the size window 
            if ind % sample == 0:
                loss_plot[ind - 1] = np.mean([ # from the start of the window to me get the average
                    self.loss_memory[i] for i in range(ind - window, ind, 1)
                ])
            ind += 1
        return loss_plot
    
    def plot_loss(self, window=20, start_ind=0, sample=1, dir='', title="Loss", 
                  TFD={}, close_me=False, 
                  ):
        """
        plot train loss for agent
        """
        loss_plot = {}
        ind = window
        while ind <= len(self.loss_memory):
            # sample
            if ind % sample == 0:
                loss_plot[ind - 1] = np.mean([
                    self.loss_memory[i] for i in range(ind - window, ind, 1)
                ])
            ind += 1
        # plot
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(8, 6))
        ax.plot(
            list(loss_plot.keys())[start_ind:],
            list(loss_plot.values())[start_ind:],
            'b-'
        )
        ax.set_xlabel('Epochs')
        ax.set_ylabel('Loss')
        ax.set_title(title, fontdict=TFD)
        fig.tight_layout()
        fig.savefig('{}{}_loss.png'.format(dir, self.name), dpi=600)
        if close_me:
            plt.close()
        return


class DRL_Env:
    """
    Deep Reinforcement learning Environment.
    `name`: str, name of this instance;
    `agent`: Agent class;
    `initial state`: tuple, terminal state = 'Delta'!!!;
    `trans_func`: input state and action,output a new state;
    `reward_func`: input state and action, output a number;
    `memory_size`: int, size of memory;
    `sample_episodes`: size of the training data

    """

    def __init__(
        self, name, agent, initial_state, trans_func, reward_func,
        memory_size=1000, sample_episodes=100, **kwargs
    ):
        # input parameters
        self.name = name
        self.agent = agent
        self.initial_state = initial_state
        self.trans_func = trans_func
        self.reward_func = reward_func
        self.learn_ind = 1
        self.memory_size = memory_size
        self.sample_episodes = sample_episodes
        # initialize memory
        self.memory = Memory(
            memory_size=self.memory_size,
            sample_episodes=self.sample_episodes
        )

    def __Q_update(
        self, discount_factor,
        learn_step, write_log
    ):
        """
        one episode of Q-learning
        """
        # =================== initialize ===================
        # initialize
        G = 0
        state = self.initial_state
        # ======================= LOOP ========================
        epoch, learn_ind = 0, 0
        # until we get to the terminal state
        while state != 'Delta':
            # take action
            action = self.agent.take_action(state)
            # transition
            new_state = self.trans_func(epoch, list(state).copy(), action)
            
            # rewards
            R = self.reward_func(state, action, new_state)
            # delta
            delta = True if new_state == 'Delta' else False
            # update G
            G += R
            
            # --------------- output experience ----------------
            if write_log:
                # logging state, action and reward
                logging.info('    epoch: {}'.format(epoch))
                logging.info('    state: {}'.format(state))
                logging.info('    action: {}'.format(action))
                logging.info('    reward: {}'.format(R))
            # ---------------- memory control --------------------
            self.memory.update(
                state, delta, action, R,
                state if new_state == "Delta" else new_state
            )
            # ---------------- Learning --------------------
            # for each agent, provide data and train.
            if learn_ind % learn_step == 0:
                self.agent.learn(
                    memory=self.memory.sample(),
                    discount_factor=discount_factor
                )
            # ---------- on step, transit to new state -------------
            state = new_state
            epoch += 1
            learn_ind += 1
        # agent record G
        self.agent.G_memory.append(G)
        # print(G)
        return G

    def deep_Q_Network(
        self, episodes, discount_factor, learn_step,
        eps_init, eps_end, write_log, report_iter=100,
    ):
        """
        Deep Q-Network.
        - episodes: how many iterations to simulation in total.
        - alpha: learning rate;
        - discount factor: perspective of future.
        """
        # return
        G = []
        print(episodes)
        # ---------------------- Learning ----------------------
        if write_log:
            logging.info("Learning...")
        for iter in range(episodes):
            if iter % report_iter == 0:
                print("Iteration {}".format(iter))
            if write_log:
                logging.info("Iteration {}".format(iter))
            step_G = self.__Q_update(
                discount_factor, learn_step, write_log
            )
            # -------------- step control ---------------
            # record return
            G.append(step_G)
            # log return
            if write_log:
                logging.info("    return: {}".format(step_G))
            # set epsilon, 7-2.5, 10-5
            self.agent.epsilon = (eps_init - eps_end) * np.max([
                (episodes * 1 - iter) / (episodes * 1), 0
            ]) + eps_end
            if write_log:
                logging.info("    -----------------------")
        return G


class Memory():
    """
    Memory: remembers past information regarding state,
        action, reward and terminal;
    """
    def __init__(self, memory_size, sample_episodes):
        super().__init__()
        self.memory_max = memory_size
        self.sample_size = sample_episodes
        self.memory = {
            'state': [], 'delta': [], 'n_state': [],
            'action': [],
            'reward': [],
        }
        self.pointer = 0
        self.memory_size = 0

    # update, keep the most recent
    def update(self, state, delta, action, reward, new_state):
        """
        `state`: list, new state, do not remember 'Delta';
        `action`: int/str, action;
        `delta`: bool, whether the NEXT state is delta;
        `reward`: double, reward.
        """
        # not full
        if self.memory_size < self.memory_max:
            self.memory['state'].append(state)
            self.memory['delta'].append(delta)
            self.memory['n_state'].append(new_state)
            self.memory['action'].append(action)
            self.memory['reward'].append(reward)
            self.memory_size += 1
        # full
        else:
            self.memory['state'][self.pointer] = state
            self.memory['delta'][self.pointer] = delta
            self.memory['n_state'][self.pointer] = new_state
            self.memory['action'][self.pointer] = action
            self.memory['reward'][self.pointer] = reward
            self.pointer += 1
        # update pointer
        if self.pointer == self.memory_max - 1:
            self.pointer = 0
        return

    # smple from the entire memory, with large memory size
    def sample(self):
        """
        sample state, action, reward and delta
        """
        # indices
        if self.memory_size >= self.sample_size:
            choose_size = self.sample_size
        else:
            choose_size = self.memory_size
        sample_ind = np.random.choice(
            range(self.memory_size), size=choose_size, replace=False,
            p=[1/self.memory_size] * self.memory_size
        )
        # sample
        state_sample = [
            self.memory['state'][i] for i in sample_ind
        ]
        new_state_sample = [
            self.memory['n_state'][i] for i in sample_ind
        ]
        delta_sample = [
            self.memory['delta'][i] for i in sample_ind
        ]
        action_sample = [
            self.memory['action'][i] for i in sample_ind
        ]
        reward_sample = [
            self.memory['reward'][i] for i in sample_ind
        ]
        return (
            state_sample, new_state_sample, delta_sample,
            action_sample, reward_sample
        )



"""
Deep Reinforcement learning.
"""

# import
# import pickle
import logging
import numpy as np
import torch
import torch.optim as optim
from ActorCritic import Actor, Critic
import matplotlib.pyplot as plt


def action_filter(state):
    """
    output valid actions for the state
    """
    actions = []
    for i in range(1, len(state)):
        if state[i] == 0:
            actions.append(i)
    if len(actions) == 0:
        actions.append(0)
    return actions

def ess_action_filter(state):
    """_summary_

    Args:
        state (_type_): represents current MG state:
                        [time, sky_condition, SOC]
        F (_type_): Max consecutive hours of discharge possible

    Returns:
        _type_: list of potential actions represented as values
                0=discharge
                1=charge
                2=idle
    """
    #                        D   C   I
    #                        0   1   2  Max hours
    # ESS directives values [1, -.c, 0, F]
    # look at the current SOC for the ESS and decide potential states
    state = list(state)
    SOC = state[2]
    F = state[-1]
    state = tuple(state)
    # if not fully discharged or fully charged anything goes
    # if 0 < SOC < F: 
    if SOC < F and SOC > 0:
        return [0, 1, 2]
    # if fully discharged we can only charge(1) or remain idle (2)
    elif SOC >= F:
        return [1]
    # if fully charged we can only discharge(1) or remain idle (0)
    elif SOC <= 0:
        return [0]
    
def abs_nrm_ess_evaluator(val, mn, mx, **kwargs):
    R = mn
    # normalize the score
    valnrm = R*((val-mn)/(mx-mn))
    return R - abs(valnrm)

def base_ess_reward(val, **kwargs):
    return val

def base_rpf_scale_ess_reward(val, rpf_scl = .2, goal=200, **kwargs):
    if val < 0:
        return val
    elif val > 0:
        return val * rpf_scl
    elif val == 0:
        return goal

def ess_state_action_reward(state, action, ess_values, PV, D, evaluator=max, 
                            verbose=False, **kwargs):
    ess_out = ess_values[action]*ess_values[-1]
    t = state[0]
    k = state[1]
    pv_out, demand = PV[t, k], D[t]
    return evaluator(PV+ess_out - demand, **kwargs)


class Agent:
    """
    Agent use in the DRL class.\n
    `name`: str, name of the agent;\n
    `actions`: a list of all actions;\n
    `input_size`: int, size of the input, len(state);\n
    `actor_hidden_layers`: list, the number of neurons for the hidden layer;\n
    `critic_hidden_layers`: list, the number of neurons for the hidden layer;\n
    `output_size`: int, size of the output, len(all actions);\n
    Keyword Arguments:\n
    `action_filter`: Boolean, True to use the action filter, False to not use;\n
    """

    def __init__(
        self, name, actions, input_size, actor_hidden_layers,
        critic_hidden_layers, output_size, actor_lr, critic_lr, **kwargs
    ):
        super().__init__()
        # ------------------- name & actions ----------------------
        self.name, self.actions = name, actions
        # ------------------- GPU ----------------
        if torch.cuda.is_available():
            self.dev = "cuda:0"
        else:
            self.dev = "cpu"
        # ------------- whether to use action_filter ----------------
        if 'action_filter' in kwargs:
                self.filter_action = True
                self.action_filter = kwargs['action_filter']
        else:        
            self.filter_action = False
        # ----------------- construct network --------------------
        # actor
        self.actor = Actor(
            hidden_layers=actor_hidden_layers,
            input_size=input_size, output_size=output_size
        )
        self.actor.to(self.dev)
        # critic
        self.critic = Critic(
            hidden_layers=critic_hidden_layers, input_size=input_size
        )
        self.critic.to(self.dev)
        # ------------------------ optimizer ----------------------------
        self.actor_optimizer = optim.Adam(
            self.actor.parameters(), lr=actor_lr
        )
        self.critic_optimizer = optim.Adam(
            self.critic.parameters(), lr=critic_lr
        )
        # self.actor_optimizer = optim.Adadelta(self.actor.parameters())
        # self.critic_optimizer = optim.Adadelta(self.critic.parameters())
        # ------------------------ training step ------------------------
        self.train_step = 1
        self.actor_loss = []
        self.critic_loss = []
        self.reward_memory = []
        self.log_prob = []
        self.V_pred = []
        self.G_memory = []
        self.entropy = 0

    def load_parameter(self, actor_path, critic_path):
        """
        Load network parameter from path
        """
        self.actor.load_state_dict(torch.load(actor_path))
        self.critic.load_state_dict(torch.load(critic_path))
        return

    def generate_policy(self, states,):
        policy = {}
        for state in states:
            ky = "-".join(  [str(s)  for s in list(state)[:-1] ]   )
            policy[ ky ] = self.take_action(state)
        return policy

    def take_action(self, state):
        """
        take action (make prediction), based on the input state.
        """
        # make state to tensor
        input_seq = torch.FloatTensor(state).to(self.dev)
        # make a prediction, with grad
        policy_dist = self.actor(input_seq)
        value = self.critic(input_seq)
        # value, distribution
        self.V_pred.append(value[0])
        dist = policy_dist.to('cpu').detach().numpy()
        # valid_actions
        if self.filter_action:
            valid_actions = self.action_filter(state)
        else:
            valid_actions = self.actions
        # choose action
        action_pr = np.array([
            np.exp(dist[i]) if self.actions[i] in valid_actions
            else 0 for i in range(len(self.actions))
        ])
        if np.sum(action_pr) == 0:
            action_pr = np.array([
                1 if self.actions[i] in valid_actions
                else 0 for i in range(len(self.actions))
            ])
        action_pr = action_pr / np.sum(action_pr)
        action_ind = np.random.choice(
            range(len(self.actions)), 1, False, action_pr
        )[0]
        # log prob
        self.log_prob.append(policy_dist.squeeze(0)[action_ind])
        # entropy
        self.entropy += -1 * np.sum(np.mean(np.exp(dist)) * dist)
        # return
        return self.actions[action_ind]

    def simulate_action(self, state):
        """
        take action (for simulation), based on the input state.
        """
        # make state to tensor
        input_seq = torch.FloatTensor(list(state.values())).to(self.dev)
        # make a prediction, without grad
        self.actor.eval()
        with torch.no_grad():
            dist = self.actor(input_seq).to('cpu').numpy()
        self.actor.train()
        # valid_actions
        if self.filter_action:
            valid_actions = self.action_filter(state)
        else:
            valid_actions = self.actions
        # choose action
        action_pr = np.array([
            np.exp(dist[i]) if self.actions[i] in valid_actions
            else 0 for i in range(len(self.actions))
        ])
        if np.sum(action_pr) == 0:
            action_pr = np.array([
                1 if self.actions[i] in valid_actions
                else 0 for i in range(len(self.actions))
            ])
        action_pr = action_pr / np.sum(action_pr)
        action_ind = np.argmax(action_pr)
        return self.actions[action_ind]

    def learn(self, discount_factor, V_future=0):
        """
        train the network
        """
        # set zero grad
        self.actor_optimizer.zero_grad()
        self.critic_optimizer.zero_grad()
        # target values
        V_targ = np.zeros(len(self.V_pred))
        for t in range(len(self.reward_memory) - 1, -1, -1):
            V_future = self.reward_memory[t] + discount_factor * V_future
            V_targ[t] = V_future
        # make tensors
        V_pred = torch.stack(self.V_pred)
        V_targ = torch.FloatTensor(V_targ).to(self.dev)
        log_prob = torch.stack(self.log_prob)
        # loss
        advantage = V_targ - V_pred
        actor_loss = (-1 * log_prob * advantage.detach()).mean()
        critic_loss = advantage.pow(2).mean()
        # back propagate
        actor_loss.backward()
        critic_loss.backward()
        self.actor_optimizer.step()
        self.critic_optimizer.step()
        # save loss
        self.actor_loss.append(actor_loss.to('cpu').detach().numpy())
        self.critic_loss.append(critic_loss.to('cpu').detach().numpy())
        # clear memory
        self.log_prob = []
        self.V_pred = []
        self.reward_memory = []
        return
    
    def plot_G(self, window=20, start_ind=0, sample=1, dir='', 
               title="G", TFD={}, close_me=False):
        """
        plot return using time window
        dir: 'directory/'
        """
        G_plot = {}
        ind = window
        while ind <= len(self.G_memory):
            # sample
            if ind % sample == 0:
                G_plot[ind - 1] = np.mean([
                    self.G_memory[i] for i in range(ind - window, ind, 1)
                ])
            ind += 1
        # plot
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(8, 6))
        ax.plot(
            list(G_plot.keys())[start_ind:],
            list(G_plot.values())[start_ind:],
            'b-'
        )
        ax.set_xlabel('Epochs')
        ax.set_ylabel('G')
        ax.set_title(title, fontdict=TFD)
        fig.tight_layout()
        fig.savefig('{}{}_G.png'.format(dir, self.name), dpi=600)
        if close_me:
            plt.close()
        return
    
    def plot_loss(self, window=20, start_ind=0, sample=1, dir='', title="Average Return", 
                  TFD={}, close_me=False):
        """
        plot train loss for agent
        dir: 'directory/'
        """
        for i in range(2):
            loss_memory = self.actor_loss if i == 0 else self.critic_loss
            loss_plot = {}
            ind = window
            while ind <= len(loss_memory):
                # sample
                if ind % sample == 0:
                    loss_plot[ind - 1] = np.mean([
                        loss_memory[i] for i in range(ind - window, ind, 1)
                    ])
                ind += 1
            # plot
            fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(8, 6))
            ax.plot(
                list(loss_plot.keys())[start_ind:],
                list(loss_plot.values())[start_ind:],
                'b-'
            )
            ax.set_xlabel('Epochs')
            ax.set_ylabel('Loss')
            ax.set_title(title, fontdict=TFD) 
            fig.tight_layout()
            fig.savefig('{}{}_{}_loss.png'.format(
                dir, self.name, "actor" if i == 0 else "critic"
            ), dpi=600)
        return


class DRL_Env:
    """
    Multi-agents Deep Reinforcement learning class.
    - initial state: dict, or functions that returns a dict,
        denoting the initial state;
        Terminal state = 'Delta'!!!!!!!!!!!!!!!!!;
    - reward: input state and action, output a number;
        Output should be a dict, {agent_name: agent_action};
    - transition: input state and action, action should be a dict,
        {agent_name: agent_action};
        If additional environment is needed, input should be
        (state, action, envs), otherwise, (state, action);
        Output a new state;
    Keyword arguments:
    - every kwargs is considered to be an agent.
    """

    def __init__(
        self, name, initial_state, trans_func, reward_func, agent
    ):
        # input parameters
        self.name = name
        self.initial_state = initial_state
        self.trans_func = trans_func
        self.reward_func = reward_func
        self.learn_step = 50
        # agent
        self.agent = agent

    def __A2C_update(self, discount_factor, write_log):
        """
        one episode of A2C
        """
        # initialize
        state = self.initial_state
        # return
        G = 0
        # loop
        epoch, V_future = 0, 0
        while state != 'Delta':
            # take action
            action = self.agent.take_action(state)
            # transition
            new_state = self.trans_func(epoch, state, action)
            # rewards
            R = self.reward_func(state, action, new_state)
            # record
            self.agent.reward_memory.append(R)
            # update G
            G += R
            # output experience
            if write_log:
                # logging state, action and reward
                logging.info('    epoch: {}'.format(epoch))
                logging.info('    state: {}'.format(state))
                logging.info('    action: {}'.format(action))
                logging.info('    reward: {}'.format(R))
            # on step, transit to new state
            state = new_state
            epoch += 1
            if epoch >= self.learn_step:
                V_future = self.agent.critic(list(state.values))
                V_future = V_future.to('cpu').detach().numpy()[0]
                break
        # agent record G
        if V_future == 0:
            self.agent.G_memory.append(G)
        # ================ Learning =================
        self.agent.learn(
            V_future=V_future,
            discount_factor=discount_factor
        )
        return G

    def advantage_actor_acritic(
        self, episodes, discount_factor=1, write_log=False
    ):
        """
        Deep Q learning.
        - episodes: how many iterations to simulation in total.
        - alpha: learning rate;
        - discount factor: perspective of future.
        """
        G = []
        # ---------------------- Learning ----------------------
        if write_log:
            logging.info("Learning...")
        for iter in range(episodes):
            if iter % 10000 == 0:
                print("Iteration {}".format(iter))
            # if True:
            if write_log:
                logging.info("Iteration {}".format(iter))
            # run one update
            # try:
            step_G = self.__A2C_update(discount_factor, write_log)
            # ---------------- step control --------------------
            # record return
            G.append(step_G)
            # log return
            if write_log:
                logging.info("    return: {}".format(step_G))
                logging.info("    -----------------------")
        return G

    def simulate(self, write_to_file, run_time):
        """
        simulate the delivery process.
        """
        if write_to_file:
            output_file = open('results/{}.txt'.format(self.name), 'w+')
            output_file.write("=============== SIMULATION ==============\n")
            output_file.write("Run Time = {}\n".format(run_time))
        # =================== initialize ===================
        # initialize
        if callable(self.initial_state):
            state = self.initial_state()
        else:
            state = self.initial_state
        # return
        G = 0
        # env
        if self.agent.use_env:
            envs = self.agent.environment
        # ======================= LOOP ========================
        epoch = 0
        while state != 'Delta':
            # ------------------ take action ------------------
            action = self.agent.simulate_action(state)
            # --------------- transit and reward ---------------
            if self.agent.use_env:
                # transition
                new_state = self.trans_func(state, action, envs)
                # rewards
                R = self.reward_func(state, action, new_state, envs)
            else:
                # transition
                new_state = self.trans_func(state, action)
                # rewards
                R = self.reward_func(state, action, new_state)
            if write_to_file:
                # logging state, action and reward
                output_file.write('    epoch: {}\n'.format(epoch))
                output_file.write('    state: {}\n'.format(state))
                output_file.write('    action: {}\n'.format(action))
                output_file.write('    reward: {}\n'.format(R))
            # ------------ exit if too many epochs --------------
            if epoch > 100:
                new_state = 'Delta'
            # -------------------- update G ----------------------
            G += R
            # ---------- on step, transit to new state -----------
            state = new_state
            epoch += 1
        if write_to_file:
            output_file.write('Return: {}\n'.format(G))
        return G



# Purpose: Classes for generating expected PV energy output  data for a month/hour combinations. 
#          The classes are used to generate data based on csv file used to generate fitted distributions 
#          used to generate the samples.
# Author: Gerald, Jones, gjones2@vols.utk.edu
# Creation Date: 2021-06-07
# Last Modified: 2024-01-07
# License: Appache 2.0

from distfit import distfit
import pandas as pd
import numpy as np

# import joblib
# import matplotlib.pyplot as plt

class MonthlyHourlyPVGenerator:
    """Can be fitted with a data set of PV output tied to a  datetime. The data times are 
       decomposed into months and hours, and each month has each hour fitted to a distribution, 
       as well as it's descriptives are stored in a dictionary keyed on month:hour: descriptive. 
       The object takes a path to a csv file with a power columne (powerCol), and datetime column (dateTimecol)
       and this file is stored as a dataframe with added hour/month columns as well as optional
       converted power columns. The data is used to fit the distributions, and generate the descriptive 
       dictionary. The data format is assumed to follow that of the NREL 5 minute PV data. 
    """
    class default_gen:
        def generate(n, **kwargs):
            return 0
    
    def __init__(self, data_path, PVCap=130e6, dateTimeCol="LocalTime", powerCol="Power(MW)", 
                 formatstr=None, convert_power_units=True, units_to_add=["kW"], 
                 **kwargs):
        self.pv_df=pd.read_csv(data_path, low_memory=False)
        self.monthly_df_dict=None
        self.dateTimeCol=dateTimeCol
        self.powerCol=powerCol
        self.PVCap=PVCap
        self.generateDateTime(self.pv_df, self.dateTimeCol, formatstr=formatstr)
        
        if convert_power_units:
            if "kW" in units_to_add:
                self.addkWFromMW(self.pv_df, mwcol=self.powerCol, newkwcol="Power(kW)")
        self.addHourFromDateTime(self.pv_df)
        self.addMonthFromDateTime(self.pv_df)
        self.generate_output_generators()
        
    def generateDateTime(self, df, time_col="LocalTime", newcol="DateTime", formatstr=None):
        df[newcol] = pd.to_datetime(df[time_col], format=formatstr)

    def addHourFromDateTime(self, df, datetimecol="DateTime", newhourcol="Hour"):    
        df[newhourcol] = [v.hour for v in df[datetimecol]]
    
    def addMonthFromDateTime(self, df, datetimecol="DateTime", newmonthcol="Month"):      
        df[newmonthcol] = [v.month for v in df[datetimecol]]

    def addkWFromMW(self, df, mwcol="Power(MW)", newkwcol="Power(kW)"):
        df[newkwcol] = df[mwcol]*1000
        
    ###################################################################################
    ####         Distribution and Stats Generator tools
    def generate_monthly_hourly_stats(self, df, powercol="Power(kW)", monthcol='Month', months=range(1, 13)):
        monthly_stats_dict = dict()
        for month in months:
            monthly_stats_dict[month] = self.generate_hourly_output_stats(df.loc[df[monthcol]==month, :], )
        self.monthly_stats_dict = monthly_stats_dict
    
    def generate_hourly_output_stats(self, df,  powercol="Power(kW)", hourcol='Hour', hours=range(0, 24)):
        results = {
            "hour":[],
            "min":[],
            "max":[],
            "mu":[],
            "std":[],
        }

        for h in hours:
            results['hour'].append(h)

            results['min'].append(df.loc[df[hourcol]==h, powercol].min())
            results['max'].append(df.loc[df[hourcol]==h, powercol].max())

            results['mu'].append(df.loc[df[hourcol]==h, powercol].mean())
            results['std'].append(df.loc[df[hourcol]==h, powercol].std())
 
        return pd.DataFrame(results)

    
    def generate_Monthly_df_dict(self, df, monthcol="Month"):
        monthly_df_dict = {}
        for m in df[monthcol]:
            monthly_df_dict[m] = df.loc[df[monthcol] == m, :]
        
        self.monthly_df_dict = monthly_df_dict
        return monthly_df_dict


    def generate_fitted_distros(self, M_df_dict, powercol="Power(kW)", hourcol='Hour', hours=range(0, 24), months=range(1, 13)):
        time = 0
        distGens = dict()
        for month in months:
            # get the dataframe for the current month from dictionary
            month_df = M_df_dict[month]
            # add distribution geneartor dictionary for each hour for this month
            distGens[month]=dict()
            for time in hours:
                # in the case the value over the hour is all zeros just set it to return 0
                # the default_gen.generate() just returns 0
                if month_df[powercol].max() == 0:
                    distGens[month][time] = self.default_gen(0)
                else:
                    distroB = month_df.loc[month_df[hourcol]==time, powercol].dropna().values.ravel()
                    dFIT = distfit()
                    res = dFIT.fit_transform(distroB)
                    distGens[month][time] = dFIT
        return distGens


    def generate_output_generators(self,):
        # store each hours data keyed on the hour (0-23)
        monthly_df_dict = self.generate_Monthly_df_dict(self.pv_df)
        self.generate_monthly_hourly_stats(self.pv_df, powercol="Power(kW)", monthcol='Month', months=range(1, 13))
        distro_gens = self.generate_fitted_distros(monthly_df_dict)
        self.distro_gens = distro_gens
        return 

    def generateOutput(self, month, hour, resamples=12, **kwargs):
        samples = list()
        mxv = self.pullStatOutput(month, hour, stat="max")
        if mxv == 0:
            return 0
        
        mnv = self.pullStatOutput(month, hour, stat="min")
        val = abs(self.distro_gens[month][hour].generate(n=1, verbose=False)[0])
        val = max(mnv, val)
        val = min(mxv, val)
        
        # for a in range(resamples): # get hours worth of 5 minute samples (60/5)
            # samples.append(abs(self.distro_gens[month][hour].generate(n=1, verbose=False))[0])
        # return np.mean(samples)
        return np.round(val, 3)
    
    def pullAvgOutput(self, month, hour):
        avg = self.monthly_stats_dict[month].loc[self.monthly_stats_dict[month]['hour']==hour, 'mu'].values[0]
        return np.around(avg, 3)
    
    def pullMinOutput(self, month, hour):
        minv = self.monthly_stats_dict[month].loc[self.monthly_stats_dict[month]['hour']==hour, 'min'].values[0]
        return np.around(minv, 3)
    
    def pullMaxOutput(self, month, hour):
        maxv = self.monthly_stats_dict[month].loc[self.monthly_stats_dict[month]['hour']==hour, 'max'].values[0]
        return np.around(maxv, 3)
    
    def pullStatOutput(self, month, hour, stat='mu'):
        v = self.monthly_stats_dict[month].loc[self.monthly_stats_dict[month]['hour']==hour, stat].values[0]
        return np.around(v, 3)

    def getPVoutDD(self, month, hour, weather, **kwargs):
        if weather == 0:
            return self.pullStatOutput(month, hour, stat="max")
        elif weather == 1:
            return self.pullStatOutput(month, hour, stat="mu")
        else:
            return self.pullStatOutput(month, hour, stat="min")






"""
MLP model, for RL.
"""

# import
import torch
import torch.nn as nn
# import torch.nn.functional as F
from collections import OrderedDict


# MLP class
class Q_MLP(nn.Module):
    """
    MLP for deep Q-learning.
    """
    def __init__(self, hidden_layer_shape, input_size, output_size, seed):
        """
        hidden_layer_shape: list, the number of neurons for every layer;
        input_size: number of states;
        output_size: number of actions;
        seed: random seed.
        """
        super().__init__()
        # parameters
        self.seed = torch.manual_seed(seed)
        self.hidden_layer_shape = hidden_layer_shape
        # NN, adding layers dinamically.
        linear = OrderedDict()
        for i in range(len(self.hidden_layer_shape) + 1):
            # input
            if i == 0:
                linear[str(i)] = nn.Linear(
                    input_size, self.hidden_layer_shape[i]
                )
            # output
            elif i == len(self.hidden_layer_shape):
                linear[str(i)] = nn.Linear(
                    self.hidden_layer_shape[i - 1], output_size
                )
            # ``hidden" layers
            else:
                linear[str(i)] = nn.Linear(
                    self.hidden_layer_shape[i - 1],
                    self.hidden_layer_shape[i]
                )
        self.layers = nn.Sequential(linear)

    def forward(self, input_seq):
        """
        input_seq: states, torch.tensor.
        """
        linear_out = 0
        # make a prediction
        for i in range(len(self.hidden_layer_shape) + 1):
            current_layer = self.layers[i]
            # input
            if i == 0:
                linear_out = torch.sigmoid(current_layer(input_seq))
            # output, no relu
            elif i == len(self.hidden_layer_shape):
                linear_out = current_layer(linear_out)
            # hidden layers
            else:
                linear_out = torch.sigmoid(current_layer(linear_out))
        # returns the prediction
        return linear_out



"""
__main__.py

This script is the main entry point of the application. It checks if a CSV file path is provided as a command line argument, 
and then checks if the file exists. The CSV file is expected to contain data for the PVGenerators module.

Author: Gerald Jones
Email: gjones2@vols.utk.edu
Creation Date: 2024-01-07
Last Edited: 2024-01-07
"""
import argparse
import json
import pandas as pd
import joblib
import time
from .PVGenerators import MonthlyHourlyPVGenerator as PVGenerator

def main():
    parser = argparse.ArgumentParser(description='Process a JSON file.')
    parser.add_argument('json_file', type=str, help='Path to the JSON file')

    args = parser.parse_args()

    with open(args.json_file, 'r') as f:
        data = json.load(f)

    for name, path in data['paths'].items():
        # df = pd.read_csv(path)
        ts = time.time()
        cap = float(name.split("MW")[0])
        print("Generating {} capacity PV generator".format(cap))
        pv_generator = PVGenerator(path, PVCap=cap,)
        joblib.dump(pv_generator, f"{data['destination']}/PVgen-{name}.sav")
        print("Time to generate PV generator: {}".format(time.time()-ts))

if __name__ == "__main__":
    main()


"""
	The file contains the PPO class to train with.
	NOTE: All "ALG STEP"s are following the numbers from the original PPO pseudocode.
			It can be found here: https://spinningup.openai.com/en/latest/_images/math/e62a8971472597f4b014c2da064f636ffe365ba3.svg
"""

import gym
import time

import numpy as np
import time
import torch
import torch.nn as nn
from torch.optim import Adam
from torch.distributions import MultivariateNormal

class PPO:
	"""
		This is the PPO modified from the original found at: 
            https://github.com/ericyangyu/PPO-for-Beginners/blob/master/ppo.py
	"""
	def __init__(self, policy_class, env, **hyperparameters):
		"""
			Initializes the PPO model, including hyperparameters.

			Parameters:
				policy_class - the policy class to use for our actor/critic networks.
				env - the environment to train on.
				hyperparameters - all extra arguments passed into PPO that should be hyperparameters.

			Returns:
				None
		"""
		# Make sure the environment is compatible with our code
		assert(type(env.observation_space) == gym.spaces.Box or type(env.observation_space) == list or type(env.observation_space) == np.array)
		assert(type(env.action_space) == gym.spaces.Box)

		# Initialize hyperparameters for training with PPO
		self._init_hyperparameters(hyperparameters)

		# Extract environment information
		self.env = env
		self.obs_dim = env.observation_space.shape[0]
		self.act_dim = env.action_space.shape[0]

		 # Initialize actor and critic networks
		self.actor = policy_class(self.obs_dim, self.act_dim)                                                   # ALG STEP 1
		self.critic = policy_class(self.obs_dim, 1)

		# Initialize optimizers for actor and critic
		self.actor_optim = Adam(self.actor.parameters(), lr=self.lr, betas=actor_betas)
		self.critic_optim = Adam(self.critic.parameters(), lr=self.lr, betas=critic_betas)

		# Initialize the covariance matrix used to query the actor for actions
		self.cov_var = torch.full(size=(self.act_dim,), fill_value=0.5)
		self.cov_mat = torch.diag(self.cov_var)

		# This logger will help us with printing out summaries of each iteration
		self.logger = {
			'delta_t': time.time_ns(),
			't_so_far': 0,          # timesteps so far
			'i_so_far': 0,          # iterations so far
			'batch_lens': [],       # episodic lengths in batch
			'batch_rews': [],       # episodic returns in batch
			'actor_losses': [],     # losses of actor network in current iteration
		}

	def learn(self, total_timesteps):
		"""
			Train the actor and critic networks. Here is where the main PPO algorithm resides.

			Parameters:
				total_timesteps := the total number of timesteps to train for

			Return:
				None
		"""
		print(f"Learning... Running {self.max_timesteps_per_episode} timesteps per episode, ", end='')
		print(f"{self.timesteps_per_batch} timesteps per batch for a total of {total_timesteps} timesteps")
		t_so_far = 0 # Timesteps simulated so far
		i_so_far = 0 # Iterations ran so far
		while t_so_far < total_timesteps:                                                                       # ALG STEP 2
			# Autobots, roll out (just kidding, we're collecting our batch simulations here)
			batch_obs, batch_acts, batch_log_probs, batch_rtgs, batch_lens = self.rollout()                     # ALG STEP 3

			# Calculate how many timesteps we collected this batch
			t_so_far += np.sum(batch_lens)

			# Increment the number of iterations
			i_so_far += 1

			# Logging timesteps so far and iterations so far
			self.logger['t_so_far'] = t_so_far
			self.logger['i_so_far'] = i_so_far

			# Calculate advantage at k-th iteration
			V, _ = self.evaluate(batch_obs, batch_acts)
			A_k = batch_rtgs - V.detach()                                                                       # ALG STEP 5

			# One of the only tricks I use that isn't in the pseudocode. Normalizing advantages
			# isn't theoretically necessary, but in practice it decreases the variance of 
			# our advantages and makes convergence much more stable and faster. I added this because
			# solving some environments was too unstable without it.
			A_k = (A_k - A_k.mean()) / (A_k.std() + 1e-10)

			# This is the loop where we update our network for some n epochs
			for _ in range(self.n_updates_per_iteration):                                                       # ALG STEP 6 & 7
				# Calculate V_phi and pi_theta(a_t | s_t)
				V, curr_log_probs = self.evaluate(batch_obs, batch_acts)

				# Calculate the ratio pi_theta(a_t | s_t) / pi_theta_k(a_t | s_t)
				# NOTE: we just subtract the logs, which is the same as
				# dividing the values and then canceling the log with e^log.
				# For why we use log probabilities instead of actual probabilities,
				# here's a great explanation: 
				# https://cs.stackexchange.com/questions/70518/why-do-we-use-the-log-in-gradient-based-reinforcement-algorithms
				# TL;DR makes gradient ascent easier behind the scenes.
				ratios = torch.exp(curr_log_probs - batch_log_probs)

				# Calculate surrogate losses.
				surr1 = ratios * A_k
				surr2 = torch.clamp(ratios, 1 - self.clip, 1 + self.clip) * A_k

				# Calculate actor and critic losses.
				# NOTE: we take the negative min of the surrogate losses because we're trying to maximize
				# the performance function, but Adam minimizes the loss. So minimizing the negative
				# performance function maximizes it.
				actor_loss = (-torch.min(surr1, surr2)).mean()
				critic_loss = nn.MSELoss()(V, batch_rtgs)

				# Calculate gradients and perform backward propagation for actor network
				self.actor_optim.zero_grad()
				actor_loss.backward(retain_graph=True)
				self.actor_optim.step()

				# Calculate gradients and perform backward propagation for critic network
				self.critic_optim.zero_grad()
				critic_loss.backward()
				self.critic_optim.step()

				# Log actor loss
				self.logger['actor_losses'].append(actor_loss.detach())

			# Print a summary of our training so far
			self._log_summary()

			# Save our model if it's time
			if i_so_far % self.save_freq == 0:
				torch.save(self.actor.state_dict(), './ppo_actor.pth')
				torch.save(self.critic.state_dict(), './ppo_critic.pth')

	def rollout(self):
		"""
			Too many transformers references, I'm sorry. This is where we collect the batch of data
			from simulation. Since this is an on-policy algorithm, we'll need to collect a fresh batch
			of data each time we iterate the actor/critic networks.

			Parameters:
				None

			Return:
				batch_obs - the observations collected this batch. Shape: (number of timesteps, dimension of observation)
				batch_acts - the actions collected this batch. Shape: (number of timesteps, dimension of action)
				batch_log_probs - the log probabilities of each action taken this batch. Shape: (number of timesteps)
				batch_rtgs - the Rewards-To-Go of each timestep in this batch. Shape: (number of timesteps)
				batch_lens - the lengths of each episode this batch. Shape: (number of episodes)
		"""
		# Batch data. For more details, check function header.
		batch_obs = []
		batch_acts = []
		batch_log_probs = []
		batch_rews = []
		batch_rtgs = []
		batch_lens = []

		# Episodic data. Keeps track of rewards per episode, will get cleared
		# upon each new episode
		ep_rews = []

		t = 0 # Keeps track of how many timesteps we've run so far this batch

		# Keep simulating until we've run more than or equal to specified timesteps per batch
		while t < self.timesteps_per_batch:
			ep_rews = [] # rewards collected per episode

			# Reset the environment. sNote that obs is short for observation. 
			obs = self.env.reset()
			done = False

			# Run an episode for a maximum of max_timesteps_per_episode timesteps
			for ep_t in range(self.max_timesteps_per_episode):
				# If render is specified, render the environment
				if self.render and (self.logger['i_so_far'] % self.render_every_i == 0) and len(batch_lens) == 0:
					self.env.render()

				t += 1 # Increment timesteps ran this batch so far

				# Track observations in this batch
				batch_obs.append(obs)

				# Calculate action and make a step in the env. 
				# Note that rew is short for reward.
				action, log_prob = self.get_action(obs)
				obs, rew, done, _ = self.env.step(action)

				# Track recent reward, action, and action log probability
				ep_rews.append(rew)
				batch_acts.append(action)
				batch_log_probs.append(log_prob)

				# If the environment tells us the episode is terminated, break
				if done:
					break

			# Track episodic lengths and rewards
			batch_lens.append(ep_t + 1)
			batch_rews.append(ep_rews)

		# Reshape data as tensors in the shape specified in function description, before returning
		batch_obs = torch.tensor(batch_obs, dtype=torch.float)
		batch_acts = torch.tensor(batch_acts, dtype=torch.float)
		batch_log_probs = torch.tensor(batch_log_probs, dtype=torch.float)
		batch_rtgs = self.compute_rtgs(batch_rews)                                                              # ALG STEP 4

		# Log the episodic returns and episodic lengths in this batch.
		self.logger['batch_rews'] = batch_rews
		self.logger['batch_lens'] = batch_lens

		return batch_obs, batch_acts, batch_log_probs, batch_rtgs, batch_lens

	def compute_rtgs(self, batch_rews):
		"""
			Compute the Reward-To-Go of each timestep in a batch given the rewards.

			Parameters:
				batch_rews - the rewards in a batch, Shape: (number of episodes, number of timesteps per episode)

			Return:
				batch_rtgs - the rewards to go, Shape: (number of timesteps in batch)
		"""
		# The rewards-to-go (rtg) per episode per batch to return.
		# The shape will be (num timesteps per episode)
		batch_rtgs = []

		# Iterate through each episode
		for ep_rews in reversed(batch_rews):

			discounted_reward = 0 # The discounted reward so far

			# Iterate through all rewards in the episode. We go backwards for smoother calculation of each
			# discounted return (think about why it would be harder starting from the beginning)
			for rew in reversed(ep_rews):
				discounted_reward = rew + discounted_reward * self.gamma
				batch_rtgs.insert(0, discounted_reward)

		# Convert the rewards-to-go into a tensor
		batch_rtgs = torch.tensor(batch_rtgs, dtype=torch.float)

		return batch_rtgs

	def get_action(self, obs):
		"""
			Queries an action from the actor network, should be called from rollout.

			Parameters:
				obs - the observation at the current timestep

			Return:
				action - the action to take, as a numpy array
				log_prob - the log probability of the selected action in the distribution
		"""
		# Query the actor network for a mean action
		mean = self.actor(obs)

		# Create a distribution with the mean action and std from the covariance matrix above.
		# For more information on how this distribution works, check out Andrew Ng's lecture on it:
		# https://www.youtube.com/watch?v=JjB58InuTqM
		dist = MultivariateNormal(mean, self.cov_mat)

		# Sample an action from the distribution
		action = dist.sample()

		# Calculate the log probability for that action
		log_prob = dist.log_prob(action)

		# Return the sampled action and the log probability of that action in our distribution
		return action.detach().numpy(), log_prob.detach()

	def evaluate(self, batch_obs, batch_acts):
		"""
			Estimate the values of each observation, and the log probs of
			each action in the most recent batch with the most recent
			iteration of the actor network. Should be called from learn.

			Parameters:
				batch_obs - the observations from the most recently collected batch as a tensor.
							Shape: (number of timesteps in batch, dimension of observation)
				batch_acts - the actions from the most recently collected batch as a tensor.
							Shape: (number of timesteps in batch, dimension of action)

			Return:
				V - the predicted values of batch_obs
				log_probs - the log probabilities of the actions taken in batch_acts given batch_obs
		"""
		# Query critic network for a value V for each batch_obs. Shape of V should be same as batch_rtgs
		V = self.critic(batch_obs).squeeze()

		# Calculate the log probabilities of batch actions using most recent actor network.
		# This segment of code is similar to that in get_action()
		mean = self.actor(batch_obs)
		dist = MultivariateNormal(mean, self.cov_mat)
		log_probs = dist.log_prob(batch_acts)

		# Return the value vector V of each observation in the batch
		# and log probabilities log_probs of each action in the batch
		return V, log_probs

	def _init_hyperparameters(self, hyperparameters):
		"""
			Initialize default and custom values for hyperparameters

			Parameters:
				hyperparameters - the extra arguments included when creating the PPO model, should only include
									hyperparameters defined below with custom values.

			Return:
				None
		"""
		# Initialize default values for hyperparameters
		# Algorithm hyperparameters
		self.timesteps_per_batch = 4800                 # Number of timesteps to run per batch
		self.max_timesteps_per_episode = 1600           # Max number of timesteps per episode
		self.n_updates_per_iteration = 5                # Number of times to update actor/critic per iteration
		self.lr = 0.005                                 # Learning rate of actor optimizer
		self.gamma = 0.95                               # Discount factor to be applied when calculating Rewards-To-Go
		self.clip = 0.2                                 # Recommended 0.2, helps define the threshold to clip the ratio during SGA

		# Miscellaneous parameters
		self.render = True                              # If we should render during rollout
		self.render_every_i = 10                        # Only render every n iterations
		self.save_freq = 10                             # How often we save in number of iterations
		self.seed = None                                # Sets the seed of our program, used for reproducibility of results

		# Change any default values to custom values for specified hyperparameters
		for param, val in hyperparameters.items():
			exec('self.' + param + ' = ' + str(val))

		# Sets the seed if specified
		if self.seed != None:
			# Check if our seed is valid first
			assert(type(self.seed) == int)

			# Set the seed 
			torch.manual_seed(self.seed)
			print(f"Successfully set seed to {self.seed}")

	def _log_summary(self):
		"""
			Print to stdout what we've logged so far in the most recent batch.

			Parameters:
				None

			Return:
				None
		"""
		# Calculate logging values. I use a few python shortcuts to calculate each value
		# without explaining since it's not too important to PPO; feel free to look it over,
		# and if you have any questions you can email me (look at bottom of README)
		delta_t = self.logger['delta_t']
		self.logger['delta_t'] = time.time_ns()
		delta_t = (self.logger['delta_t'] - delta_t) / 1e9
		delta_t = str(round(delta_t, 2))

		t_so_far = self.logger['t_so_far']
		i_so_far = self.logger['i_so_far']
		avg_ep_lens = np.mean(self.logger['batch_lens'])
		avg_ep_rews = np.mean([np.sum(ep_rews) for ep_rews in self.logger['batch_rews']])
		avg_actor_loss = np.mean([losses.float().mean() for losses in self.logger['actor_losses']])

		# Round decimal places for more aesthetic logging messages
		avg_ep_lens = str(round(avg_ep_lens, 2))
		avg_ep_rews = str(round(avg_ep_rews, 2))
		avg_actor_loss = str(round(avg_actor_loss, 5))

		# Print logging statements
		print(flush=True)
		print(f"-------------------- Iteration #{i_so_far} --------------------", flush=True)
		print(f"Average Episodic Length: {avg_ep_lens}", flush=True)
		print(f"Average Episodic Return: {avg_ep_rews}", flush=True)
		print(f"Average Loss: {avg_actor_loss}", flush=True)
		print(f"Timesteps So Far: {t_so_far}", flush=True)
		print(f"Iteration took: {delta_t} secs", flush=True)
		print(f"------------------------------------------------------", flush=True)
		print(flush=True)

		# Reset batch-specific logging data
		self.logger['batch_lens'] = []
		self.logger['batch_rews'] = []
		self.logger['actor_losses'] = []


"""
__main__.py

This script is the main entry point of the DemandGenerators application. It takes a path to a JSON file as a command-line argument. 
The JSON file is expected to contain data paths and a name for the Demand_Generators.DataGenerators() object.

The script does the following:
1. Parses the command-line argument to get the JSON file path.
2. Opens and loads the JSON file into a dictionary.
3. Uses the dictionary to create a Demand_Generators.DataGenerators() object.
4. Stores the Demand_Generators.DataGenerators() object using joblib. The name of the joblib file is taken from the "name" key in the JSON file.
"""
import argparse
import json
import joblib
from .DemandGenerators import DemandNodeSimulator as DataGenerators

def main():
    parser = argparse.ArgumentParser(description='Process a JSON file.')
    parser.add_argument('json_file', type=str, help='Path to the JSON file')

    args = parser.parse_args()

    with open(args.json_file, 'r') as f:
        data = json.load(f)

    demand_generator = DataGenerators(data['paths'])

    joblib.dump(demand_generator, f"{data['name']}.sav")

if __name__ == "__main__":
    main()


"""
IE-608 DQN example.
"""

# import
import logging
import numpy as np
from DRL import Agent, DRL_Env, ess_action_filter
from matplotlib import pyplot as plt


def main():
    """
    main
    """
    # --------- parameters -----------

    # horizon
    T = 16   # hours we need to cover
    K = 1    # number of potential sky conditions
    soc = 4  # number of consecutive discharge hours

    # ------------ MDP --------------

    # initial state: t, condition
    initial_state = (0, 0, 0, soc)

    # number of actions
    # actions = range(num_plans)
    actions = [0, 1, 2]
    epsi = .1              # percentage 
    action_weights = [1, -epsi, 0]    # used to control ESS supply effect

    ##################################
    #### Set up test environment
    ##################################
    seed = 13
    np.random.seed(seed)
   

    demand_min = 10
    demand_max = 50
    demands = np.random.default_rng().choice(list(np.arange(demand_min, demand_max+10, 10)), size=T)

    # set ouput max below demands to make sure there are cased when we will not meet demand
    outputs =  np.random.default_rng().choice(list(np.arange(demand_min, demand_max, T)), size=T)

    unmet = outputs - demands
    
    t = list(range(1, T+1))


    # use max unmet to set ESS output for testing
    ess_out = max(np.abs(unmet))*.80
    ess_out = max(np.abs(unmet))*2
    ess_unmet = outputs+ess_out - demands
    print(ess_out)
    
    ##################################
    ##################################
     # make some quick evaluators
    abs_reward = lambda x: -abs(x)  # just use abs val
    
    def fulfillment_specialist(x):
        if x < 0:
            return x
        return 0
    
    def fulfillment_specialist2(x):
        if x < 0:
            return x*under_weight
        return x*over_weight
    
    def linear(x):
        return x
    
    def recursive_average(new, x, t):
        if t == 0:
            return x
        return x + (1/t)(new-x)
        
    
    def trans_funcESS(t, s, a):
        """_summary_

        Args:
            t: epoch
            s: rest of state: [h,k,soc, F]
            a: [0(discharge), 1(charge), 2(idle)]
        """
        s = list(s).copy()
        if s[0] + 1 >= T:
            return "Delta"
        
        # unpack the state
        s[0] += 1 
        k = s[1]
        # sc = s[2]
        F = s[-1]
        
        pv_out = outputs[t]
        dem = demands[t]   
        
        if a == 0:
            # increment up to max value
            # s[2] = min(s[2] + 1, s[-1])
            s[2] =s[2] + 1
        elif a == 1:
            # s[2] = max(s[2] - 1, 0)
            s[2] = s[2] - 1
        
        return s
            
    def store_sample(samplev, storageV):
        storageV.append(np.mean(samplev))
            
    
    svals = []
    savgs = []
    sample = 500

    # def reward_funcESS(s, a, s_next, base_eval=linear, #):
    # def reward_funcESS(s, a, s_next, base_eval=fulfillment_specialist, #):
    # def reward_funcESS(s, a, s_next, base_eval=fulfillment_specialist2, #):
    def reward_funcESS(s, a, s_next, base_eval=abs_reward, 
                    #    ess_logic=True):
                       ess_logic=True):
        """_summary_

        Args:
            s: current state tuple (t, k, soc, F)
            a: action to be taken one of
            s_next:   state the given action would leave the system in
        Returns:
            _type_: some numeric value derived by base_eval function 
                    and current difference between DG output and Demand
        """
        s = list(s)
        t = s[0]       # current time
        k = s[1]
        sc = s[2]
        F = s[-1]      # get max SOC
        pv_out = outputs[t]
        dem = demands[t]
        
        if ess_logic:
            # punish bad actions
            if (sc >= F) and a==0:  # punish discharge when empty
                return -1e1
            if (sc <=0) and a == 1: # punish full charging
                return -1e1
            
            # punish inaction
            if (sc == F) and a == 2: # punish not charging on empty
                return -1e1
            if (sc == 0) and a == 2:  # not discharge when full
                return -1e1
        
        # use action to determine the ESS output 
        # also based on SOC and action filter used
        if a == 0 or a == 2:
            ess_con = action_weights[a]*ess_out
        else:
            # ess_con = action_weights[a]*pv_out
            ess_con = -.1*pv_out
        
        O = pv_out + ess_con 
        S = O - dem
        
        # if len(svals) > 2:
        #     # normalize?
        #     S = (S-np.min(svals))/(np.max(svals) - np.min(svals))
        
        svals.append(S)
        if len(svals)%sample == 0 and t > 0:
            # print(len(svals))
            store_sample(svals, savgs)
            svals.clear()
        
        
        return base_eval(S)


    def reward_funcESSOG(s, a, s_next, base_eval=abs_reward):
        """_summary_

        Args:
            s: current state tuple (t, k, soc, F)
            a: action to be taken one of
            s_next:   state the given action would leave the system in
        Returns:
            _type_: some numeric value derived by base_eval function 
                    and current difference between DG output and Demand
        """
        s = list(s)
        t = s[0]       # current time
        k = s[1]
        F = s[-1]      # get max SOC
        pv_out = outputs[t]
        dem = demands[t]
        
        # use action to determine the ESS output 
        # also based on SOC and action filter used
        ess_con = action_weights[a]*ess_out
        
        S = pv_out + ess_con - dem
        
        return base_eval(S)
             

    # learning params
    
    ##############################3################
    #####                Agent  params
    ##############################3################
    episodes=1200
    hidden_layers = [100, 100]
    lr = 1e-4 
    lr = 1e-4 
    learn_epoch = 1
    epsilon = .1   
    # agent
    manager = Agent(
        name="ESS_manager",
        actions=actions,
        input_size=len(initial_state),
        hidden_layers=hidden_layers,
        output_size=len(actions),
        learning_rate=lr, 
        learn_epoch=learn_epoch, 
        epsilon=epsilon,
        action_filter=ess_action_filter,
    )

    # define problem
    ################################### 
    # create the problem environment 
    # this requires the agent, the transition logic function, and the reward function
    mem_size = 5000
    samples_ep = 500
    
    problem = DRL_Env(
        name="maintenance",
        agent=manager,
        initial_state=initial_state,  # start at t=1
        trans_func=trans_funcESS,
        reward_func=reward_funcESS,
        memory_size=mem_size,
        sample_episodes=samples_ep,
    )

    ###################################################
    ##############       logging
    ###################################################
    logging.basicConfig(
        filename='maintenanceESS.log', filemode='w+',
        format='%(levelname)s - %(message)s', level=logging.INFO
    )

    ################################################
    ###########              DQN
    ################################################
    DQ_episodes = 1200
    gamma = .9
    DG_learn_step = 1
    eps_init = 1
    eps_end= 0
    write_log = False
    
    # call the Q network to optimze the problem
    G = problem.deep_Q_Network(
        episodes=DQ_episodes,
        discount_factor=gamma,
        learn_step=DG_learn_step,
        eps_init=eps_init,
        eps_end=eps_end,
        write_log=write_log,
    )


    
    # plot
    manager.plot_loss(window=20, start_ind=0, sample=1, dir='./figs',
                      title="My LOSS", TFD={"size": 20})
    # plt.show()
    
    manager.plot_G(window=20, start_ind=0, sample=1, dir='./figs',
                      title="Expected Return", TFD={"size": 20})
    plt.show()

    def generate_all_states(T, K, soc):
        states = list()
        for t in range(T):
            for k in K:
                for sc in range(soc+1):
                    states.append([t, k, sc, soc])
        return states
    
    def generate_quick_run(T, soc):
        states = list()
        for t in range(T):
            # for sc in range(soc+1):
            states.append([t, 0, 0, soc])
        return states
    
    statesAll = generate_all_states(T, K=[0], soc=soc)
    select_states = generate_quick_run(T, soc)
    
    # show policy?
    policy = problem.agent.generate_policy(states=statesAll)
    check_policy = problem.agent.generate_policy(states=select_states)
    sky_conds = {
        0:"sunny",
        1:'cloudy',
        2: 'overcast',
    }
    
    ESSM_Action_labels = {
        0: "discharge",
        1: "charge",
        2: "Idle",
    }
    xticklabels = list()
    c = 0
    print(len(select_states))

    for s in select_states:
        print(s)

    for s in select_states:
        s[2] = c
        s = "-".join([str(sv) for sv in s[:-1]])
        sp = s.split("-")
        t = int(sp[0]) + 1
        k = sky_conds[int(sp[1])]
        soc = int(sp[2])
        a = policy[s]
        Action = ESSM_Action_labels[a]
        if a == 0: # discharge
            c += 1
        elif a == 1: # charge
            c -= 1
        xticklabels.append(Action[0].upper())
        print("------------------------------------------------------------")
        print("State: time:{}\nsky condition: {}\nSOC: {}".format(t, k, c))
        print("\t\tAction: {}".format(Action))
        print("------------------------------------------------------------\n")
    
                
    fig, ax = plt.subplots(1)
    print("ESS-out: ", ess_out)
    print(demands)
    print(outputs)
    xl = range(len(outputs))
    ax.bar(xl, demands, color="red", label="Unmet")
    ax.bar(xl, outputs, color='green', alpha=.5, label="Exceeds")
    ax.bar(1, 0, color="brown", label="Meet")
    ax.set_xticks(xl)
    ax.set_xticklabels(xticklabels, fontdict={"rotation":90})
    plt.legend()
    plt.show()

    fig, ax = plt.subplots(1)
    
    print(savgs)
    ax.plot(savgs, label="averaged Supplied values")
    # ax.set_xticks(xl)
    # ax.set_xticklabels(xticklabels, fontdict={"rotation":90})
    plt.legend()
    plt.show()
    return

if __name__ == "__main__":
    print("Hi, it's a me, Jonesy!")
    main()



"""
MLP model, for RL.
"""

# import
import logging
import torch
import torch.nn as nn


class ActorCritic(nn.Module):
    """
    Actor Critic
    """
    def __init__(self, num_inputs, num_actions, hidden_size):
        """
        Initialize
        """
        super(ActorCritic, self).__init__()
        self.seed = torch.manual_seed(1)
        # critic
        self.num_actions = num_actions
        self.critic_linear1 = nn.Linear(num_inputs, 500)
        self.critic_actvtn1 = nn.Tanh()
        self.critic_linear2 = nn.Linear(500, 200)
        self.critic_actvtn2 = nn.Tanh()
        self.critic_linear3 = nn.Linear(200, 1)
        # actor
        self.actor_linear1 = nn.Linear(num_inputs, hidden_size)
        self.actor_actvtn1 = nn.Tanh()
        self.actor_linear2 = nn.Linear(hidden_size, num_actions)
        self.actor_actvtn2 = nn.Softmax(dim=0)

    def forward(self, input_seq):
        """
        input_seq: states, torch.tensor.
        """
        # critic
        value = self.critic_linear1(input_seq)
        value = self.critic_actvtn1(value)
        value = self.critic_linear2(value)
        value = self.critic_actvtn2(value)
        value = self.critic_linear3(value)
        # actor
        policy_dist = self.actor_linear1(input_seq)
        policy_dist = self.actor_actvtn1(policy_dist)
        policy_dist = self.actor_linear2(policy_dist)
        policy_dist = self.actor_actvtn2(policy_dist)
        # return
        return value, policy_dist


def printgradnorm_1(self, grad_input, grad_output, **kwargs):
    logging.info('Inside ' + self.__class__.__name__ + ' backward\n')
    logging.info('Inside class:' + self.__class__.__name__ + '\n')
    logging.info('grad_input: {}\n'.format(grad_input[0]))
    logging.info('grad_input size: {}\n'.format(grad_input[0].size()))
    logging.info('grad_output: {}\n'.format(grad_output))
    logging.info('grad_output size: {}\n'.format(grad_output[0].size()))


def printgradnorm_2(self, grad_input, grad_output, **kwargs):
    logging.info('Inside ' + self.__class__.__name__ + ' backward\n')
    logging.info('Inside class:' + self.__class__.__name__ + '\n')
    logging.info('grad_input: {}\n'.format(grad_input[0]))
    logging.info('grad_input size: {}\n'.format(grad_input[0].size()))
    logging.info('grad_output: {}\n'.format(grad_output))
    logging.info('grad_output size: {}\n'.format(grad_output[0].size()))


class Actor(nn.Module):
    """
    Actor network for A2C.
    """
    def __init__(self, hidden_layers, input_size, output_size, seed=1):
        """
        `hidden_layers`: list, the number of neurons for every layer;
        `input_size`: number of states;
        `output_size`: number of actions;
        `seed`: random seed.
        """
        super().__init__()
        # parameters
        self.seed = torch.manual_seed(seed)
        # NN, adding layers dynamically.
        self.layers = nn.Sequential()
        # ---------------------- input -----------------------
        self.layers.add_module(
            'Linear_inp', nn.Linear(input_size, hidden_layers[0])
        )
        self.layers.add_module('Act_inp', nn.Tanh())
        # ---------------------- hidden ----------------------
        for i in range(1, len(hidden_layers)):
            self.layers.add_module(
                'Linear_{}'.format(i),
                nn.Linear(hidden_layers[i - 1], hidden_layers[i])
            )
            self.layers.add_module('Act_{}'.format(i), nn.Tanh())
        # ----------------------- output ---------------------
        self.layers.add_module(
            'Linear_out', nn.Linear(hidden_layers[-1], output_size)
        )
        self.layers.add_module('Act_out', nn.LogSoftmax(dim=0))
        # actor
        # self.actor_linear1 = nn.Linear(input_size, hidden_layers[0])
        # self.actor_actvtn1 = nn.Tanh()
        # self.actor_linear2 = nn.Linear(hidden_layers[0], output_size)
        # self.actor_actvtn2 = nn.Sigmoid()
        # self.actor_actvtn2 = nn.LogSoftmax(dim=0)
        # self.actor_linear1.register_backward_hook(printgradnorm_1)
        # self.actor_linear2.register_backward_hook(printgradnorm_2)

    def forward(self, input_seq):
        """
        `input_seq`: states, torch.FloatTensor.
        """
        # actor
        # policy_dist = self.actor_linear1(input_seq)
        # policy_dist = self.actor_actvtn1(policy_dist)
        # policy_dist = self.actor_linear2(policy_dist)
        # policy_dist = self.actor_actvtn2(policy_dist)
        # policy_dist = self.actor_actvtn3(policy_dist)
        # return
        # return policy_dist
        return self.layers(input_seq)


class Critic(nn.Module):
    """
    Critic network for A2C.
    """
    def __init__(self, hidden_layers, input_size, seed=1):
        """
        `hidden_layers`: list, the number of neurons for every layer;
        `input_size`: number of states;
        `output_size`: number of actions;
        `seed`: random seed.
        """
        super().__init__()
        # parameters
        self.seed = torch.manual_seed(seed)
        # NN, adding layers dynamically.
        self.layers = nn.Sequential()
        # ---------------------- input -----------------------
        self.layers.add_module(
            'Linear_inp', nn.Linear(input_size, hidden_layers[0])
        )
        self.layers.add_module('Act_inp', nn.ReLU())
        # ---------------------- hidden ----------------------
        for i in range(1, len(hidden_layers)):
            self.layers.add_module(
                'Linear_{}'.format(i),
                nn.Linear(hidden_layers[i - 1], hidden_layers[i])
            )
            self.layers.add_module('Act_{}'.format(i), nn.ReLU())
        # ----------------------- output ---------------------
        self.layers.add_module(
            'Linear_out', nn.Linear(hidden_layers[-1], 1)
        )

    def forward(self, input_seq):
        """
        `input_seq`: states, torch.FloatTensor.
        """
        return self.layers(input_seq)



# Purpose: Tools to generate data from some distribution. Classes for
#          generating demand data for a given building type, and time 
#          of day are defined. In addition a class for generating PV 
#          energy outputs are defined. Lastly, a stored instances of python objects 
#          for generating demand and PV data are created and stored as variables.
# Author: Gerald, Jones, gjones2@vols.utk.edu
# Creation Date: 2021-06-07
# Last Modified: 2024-01-07
# License: Appache 2.0

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

import joblib

import os 

from distfit import distfit

class DemandNodeSimulator:
    """Can be fitted with a data set of demand data tied to a  datetime and location type. The data times are
       decomposed into hours, and each hour has each hour fitted to a distribution for all locations in the given datasets,   
       as well as it's descriptives are stored in a dictionary keyed on hour. By passing a hour and location type, a sample 
       hourly demand for the given hour and location type can be generated. The class takes a dictionary keyed on location 
       types with values of paths to the given locations demand data. The data files are expected to be csv file that have 
       columns of date/time and demand. The data/time values are converted to 24 values and used to fit distributions for 
       each hour and generate the descriptive data. 
    """
    def __init__(self, type_path_dict, seed=None, **kwargs):
        """Generates a demand generator for each of the buildings pointed to by the path dictionary

        Args:
            type_path_dict (dict): key=building type, value=path to csv file with hourly demand data. 
            seed (float, optional): seed to pass to random number generator. Defaults to None.
        """
        self.seed=seed
        self.Demand_Generators = dict()
        self.type_path_dict = type_path_dict
        self.typeList = list(type_path_dict.keys())
        self.hourly_demand_stats = dict()
        self.paths = list(type_path_dict.values())
        self.initialize()
    
    
    def initialize(self, ):
        """generates a demand generator for each of the buildings pointed to by the path dictionary
           These will be demand distributions based on the given building type, and time of day
        """
        for ntype in self.type_path_dict:
            hourly_df_dict, demand_stats, distro_gens = self.generate_demand_generators(self.type_path_dict[ntype])
            self.Demand_Generators[ntype] = distro_gens
            self.hourly_demand_stats[ntype]  = demand_stats 
    
    
    def get_hourly_total(self, hour, node_count_dict=None):
        """Calculates the total hourly demand for all buildings for the given hour. 
           The node count can be used to simulate multiple buildings of the same type.

        Args:
            hour (int): [0,..., 23] hour of the day to get the demand for
            node_count_dict (dict, optional): key=demand node type, value=number of given type. Defaults to None.

        Returns:
            float: total demand for the current hour for the current demand node structure
        """
        ## calculates the total demand for the given hour for all buildings
        demands = 0
        total_max = 0
        
        # if need to generate a dict of node counts for each node type
        if node_count_dict is None:
            node_count_dict={}
            for node in self.Demand_Generators:
                node_count_dict[node]=1
        
        # node represents the type of demand node i.e., hospital, police station, etc.
        for node in self.Demand_Generators:
            # get the data frame for the node type and hour
            hrdf =  self.hourly_demand_stats[node]
            
            # get the min/max values for the hour
            minval = hrdf.loc[hrdf['hour']==hour, 'min'].values[0]
            maxval = hrdf.loc[hrdf['hour']==hour, 'max'].values[0]
            total_max += maxval
            
            # for the number of nodes of this type generate the demand for the hour
            # while ensuring the demand is within the min/max bounds for the data
            for c in range(node_count_dict[node]):
                # ensure demand is inside min/max bounds for data
                demand_sampleHc = np.sum(self.Demand_Generators[node][hour].generate(1, verbose=False)) 
                dem = max(minval, demand_sampleHc)
                dem = min(maxval, dem)
                demands += dem
        
        return demands
            
    
    def adjust_date_col(self, df, datecol="Date/Time", ):
        df[datecol] = [str(s.strip()) for s in df[datecol]]
        dl = list()
        for d in df[datecol]:
            s = d.split("  ")
            t = s[-1]
            f = s[0].strip()
            H = t.split(':')[0]
            d = f +"/22 "+ t

            if int(H) <= 10:
                rep = "0" + str(int(H)-1)
            else:
                rep = str(int(H)-1)
            d = d.replace(H+":", rep + ":")
            dl.append(d)
        return dl

    def generate_hourly_demand_stats(self, df_dict,  hourcol='Electricity:Facility [kW](Hourly)', hours=range(0, 24), Ftype="Hospital"):
        """Generates the descriptive statistics for the given demand data
        
        
        Args:
        
            df_dict (dict): key=hour, value=dataframe for the given hour
            hourcol (str, optional): column name for the demand data. Defaults to 'Electricity:Facility [kW](Hourly)'.
            hours (list, optional): list of hours to generate the stats for. Defaults to range(0, 24).
            Ftype (str, optional): building type. Defaults to "Hospital".
            
            returns:
            
            pd.DataFrame: dataframe with the descriptive statistics for the given demand data for the given hours
            
            """
        results = {
            "hour":[],
            "min":[],
            "max":[],
            "mu":[],
            "std":[],
        }

        # for each hour generate the stats
        for h in hours:
            results['hour'].append(h)

            results['min'].append(df_dict[h][hourcol].min())
            results['max'].append(df_dict[h][hourcol].max())

            results['mu'].append(df_dict[h][hourcol].mean())
            results['std'].append(df_dict[h][hourcol].std())

        return pd.DataFrame(results)

    
    def generate_hourly_df_dict(self, df, hourcol="hour", metriccol='Electricity:Facility [kW](Hourly)', percentile=.96):
        """Generates a dictionary of dataframes for each hour of the day where the energy metric is <= than the given percentile

        Args:
            df (_type_): a dataframe with a column for the hour of the day, and a column for the energy metric
            hourcol (str, optional): column for the hour value. Defaults to "hour".
            metriccol (str, optional): _description_. Defaults to 'Electricity:Facility [kW](Hourly)'.

        Returns:
            dict: key=hour, value=dataframe for the given hour with the energy metric <= the given percentile
        """
        hourly_df_dict = {}
        for f in df[hourcol]:
            hourly_df_dict[f] = df.loc[df[hourcol] == f, :]
            hourly_df_dict[f] = hourly_df_dict[f].loc[hourly_df_dict[f][metriccol] <= hourly_df_dict[f][metriccol].quantile(percentile), :]
        return hourly_df_dict


    def generate_fitted_given_hours_distros(self, H_df_dict, hourcol='Electricity:Facility [kW](Hourly)', hours=range(0, 24), Ftype="Hospital",
                                            verbose=False):
        """Generates a dictionary of distribution generators for each hour of the day for the given demand data, relies on the distfit package
        
        Args:
        
            H_df_dict (dict): key=hour, value=dataframe for the given hour
            hourcol (str, optional): column name for the demand data. Defaults to 'Electricity:Facility [kW](Hourly)'.
            hours (list, optional): list of hours to generate the stats for. Defaults to range(0, 24).
            Ftype (str, optional): building type. Defaults to "Hospital".
            
            returns:
            
            dict: key=hour, value=distribution generator for the given hour"""
        time = 0
        distGens = dict()
        for time in hours:
            distroB = H_df_dict[time][[hourcol]].dropna().values.ravel()
            dFIT = distfit()
            res = dFIT.fit_transform(distroB)
            if verbose:
                _, ax = dFIT.plot()
                ax.set_title(f"Hour: {time}")
            distGens[time] = dFIT
        return distGens

    

    def generate_demand_generators(self, data_path):
        """
            data_path: path to csv file with columns for date/time and kw used that hour for a given building
            Expected columns: Date/Time, Electricity:Facility [kW](Hourly)
        """
        _df = pd.read_csv(data_path)   # load demand data for the building and adjust data/get hours
        # convert date/time column to string and adjust the date/time to be the hours
        _df["Date/Time"] = [str(s).strip() for s in _df["Date/Time"]]
        _df["Date/Time"] = self.adjust_date_col(_df, datecol="Date/Time", )
        _df["Date/Time"] = pd.to_datetime(_df["Date/Time"], format="%m/%d/%y %H:%M:%S") 
        _df["hour"] = _df["Date/Time"].dt.hour

        # store each hours data keyed on the hour (0-23)
        hourly_df_dict = self.generate_hourly_df_dict(_df, hourcol="hour")

        #  Take a look at the stats for each hour
        demand_stats = self.generate_hourly_demand_stats(hourly_df_dict)
        distro_gens = self.generate_fitted_given_hours_distros(hourly_df_dict)


        return hourly_df_dict, demand_stats, distro_gens 


class MonthlyHourlyPVGenerator:
    """Can be fitted with a data set of PV output tied to a  datetime. The data times are 
       decomposed into months and hours, and each month has each hour fitted to a distribution, 
       as well as it's descriptives are stored in a dictionary keyed on month:hour: descriptive. 
       The object takes a path to a csv file with a power columne (powerCol), and datetime column (dateTimecol)
       and this file is stored as a dataframe with added hour/month columns as well as optional
       converted power columns. The data is used to fit the distributions, and generate the descriptive 
       dictionary. The data format is assumed to follow that of the NREL 5 minute PV data. 
    """
    class default_gen:
        def generate(n, **kwargs):
            return 0
    
    def __init__(self, data_path, PVCap=130e6, dateTimeCol="LocalTime", powerCol="Power(MW)", 
                 formatstr=None, convert_power_units=True, units_to_add=["kW"], 
                 **kwargs):
        self.pv_df=pd.read_csv(data_path, low_memory=False)
        self.monthly_df_dict=None
        self.dateTimeCol=dateTimeCol
        self.powerCol=powerCol
        self.PVCap=PVCap
        self.generateDateTime(self.pv_df, self.dateTimeCol, formatstr=formatstr)
        
        if convert_power_units:
            if "kW" in units_to_add:
                self.addkWFromMW(self.pv_df, mwcol=self.powerCol, newkwcol="Power(kW)")
        self.addHourFromDateTime(self.pv_df)
        self.addMonthFromDateTime(self.pv_df)
        self.generate_output_generators()
        
    def generateDateTime(self, df, time_col="LocalTime", newcol="DateTime", formatstr=None):
        df[newcol] = pd.to_datetime(df[time_col], format=formatstr)

    def addHourFromDateTime(self, df, datetimecol="DateTime", newhourcol="Hour"):    
        df[newhourcol] = [v.hour for v in df[datetimecol]]
    
    def addMonthFromDateTime(self, df, datetimecol="DateTime", newmonthcol="Month"):      
        df[newmonthcol] = [v.month for v in df[datetimecol]]

    def addkWFromMW(self, df, mwcol="Power(MW)", newkwcol="Power(kW)"):
        df[newkwcol] = df[mwcol]*1000
        
    ###################################################################################
    ####         Distribution and Stats Generator tools
    def generate_monthly_hourly_stats(self, df, powercol="Power(kW)", monthcol='Month', months=range(1, 13)):
        monthly_stats_dict = dict()
        for month in months:
            monthly_stats_dict[month] = self.generate_hourly_output_stats(df.loc[df[monthcol]==month, :], )
        self.monthly_stats_dict = monthly_stats_dict
    
    def generate_hourly_output_stats(self, df,  powercol="Power(kW)", hourcol='Hour', hours=range(0, 24)):
        results = {
            "hour":[],
            "min":[],
            "max":[],
            "mu":[],
            "std":[],
        }

        for h in hours:
            results['hour'].append(h)

            results['min'].append(df.loc[df[hourcol]==h, powercol].min())
            results['max'].append(df.loc[df[hourcol]==h, powercol].max())

            results['mu'].append(df.loc[df[hourcol]==h, powercol].mean())
            results['std'].append(df.loc[df[hourcol]==h, powercol].std())
 
        return pd.DataFrame(results)

    
    def generate_Monthly_df_dict(self, df, monthcol="Month"):
        monthly_df_dict = {}
        for m in df[monthcol]:
            monthly_df_dict[m] = df.loc[df[monthcol] == m, :]
        
        self.monthly_df_dict = monthly_df_dict
        return monthly_df_dict


    def generate_fitted_distros(self, M_df_dict, powercol="Power(kW)", hourcol='Hour', hours=range(0, 24), months=range(1, 13)):
        time = 0
        distGens = dict()
        for month in months:
            # get the dataframe for the current month from dictionary
            month_df = M_df_dict[month]
            # add distribution geneartor dictionary for each hour for this month
            distGens[month]=dict()
            for time in hours:
                # in the case the value over the hour is all zeros just set it to return 0
                # the default_gen.generate() just returns 0
                if month_df[powercol].max() == 0:
                    distGens[month][time] = self.default_gen(0)
                else:
                    distroB = month_df.loc[month_df[hourcol]==time, powercol].dropna().values.ravel()
                    dFIT = distfit()
                    res = dFIT.fit_transform(distroB)
                    distGens[month][time] = dFIT
        return distGens


    def generate_output_generators(self,):
        # store each hours data keyed on the hour (0-23)
        monthly_df_dict = self.generate_Monthly_df_dict(self.pv_df)
        self.generate_monthly_hourly_stats(self.pv_df, powercol="Power(kW)", monthcol='Month', months=range(1, 13))
        distro_gens = self.generate_fitted_distros(monthly_df_dict)
        self.distro_gens = distro_gens
        return 

    def generateOutput(self, month, hour, resamples=12, **kwargs):
        samples = list()
        mxv = self.pullStatOutput(month, hour, stat="max")
        if mxv == 0:
            return 0
        
        mnv = self.pullStatOutput(month, hour, stat="min")
        val = abs(self.distro_gens[month][hour].generate(n=1, verbose=False)[0])
        val = max(mnv, val)
        val = min(mxv, val)
        
        # for a in range(resamples): # get hours worth of 5 minute samples (60/5)
            # samples.append(abs(self.distro_gens[month][hour].generate(n=1, verbose=False))[0])
        # return np.mean(samples)
        return np.round(val, 3)
    
    def pullAvgOutput(self, month, hour):
        avg = self.monthly_stats_dict[month].loc[self.monthly_stats_dict[month]['hour']==hour, 'mu'].values[0]
        return np.around(avg, 3)
    
    def pullMinOutput(self, month, hour):
        minv = self.monthly_stats_dict[month].loc[self.monthly_stats_dict[month]['hour']==hour, 'min'].values[0]
        return np.around(minv, 3)
    
    def pullMaxOutput(self, month, hour):
        maxv = self.monthly_stats_dict[month].loc[self.monthly_stats_dict[month]['hour']==hour, 'max'].values[0]
        return np.around(maxv, 3)
    
    def pullStatOutput(self, month, hour, stat='mu'):
        v = self.monthly_stats_dict[month].loc[self.monthly_stats_dict[month]['hour']==hour, stat].values[0]
        return np.around(v, 3)

    def getPVoutDD(self, month, hour, weather, **kwargs):
        if weather == 0:
            return self.pullStatOutput(month, hour, stat="max")
        elif weather == 1:
            return self.pullStatOutput(month, hour, stat="mu")
        else:
            return self.pullStatOutput(month, hour, stat="min")


class DemandNodeSimulatorBU:
    def __init__(self, type_path_dict, **kwargs):
        self.Demand_Generators = dict()
        self.type_path_dict = type_path_dict
        self.typeList = list(type_path_dict.keys())
        self.paths = list(type_path_dict.values())
        self.initialize()
    
    
    def initialize(self, ):
        for h in self.type_path_dict:
            hourly_df_dict, demand_stats, distro_gens = self.generate_demand_generators(self.type_path_dict[h])
            self.Demand_Generators[h] = distro_gens
    
    def get_hourly_total(self, hour, node_count_dict=None):
        demands = 0
        if node_count_dict is None:
            node_count_dict={}
            for node in self.Demand_Generators:
                node_count_dict[node]=1
        for node in self.Demand_Generators:
            demands += np.sum(self.Demand_Generators[node][hour].generate(node_count_dict[node], verbose=False))
        return demands
            
    
    def adjust_date_col(self, df, datecol="Date/Time", ):
        df[datecol] = [str(s.strip()) for s in df[datecol]]
        dl = list()
        for d in df[datecol]:
            s = d.split("  ")
            t = s[-1]
            f = s[0].strip()
            H = t.split(':')[0]
            d = f +"/22 "+ t

            if int(H) <= 10:
                rep = "0" + str(int(H)-1)
            else:
                rep = str(int(H)-1)
            d = d.replace(H+":", rep + ":")
            dl.append(d)
        return dl

    def generate_hourly_demand_stats(self, df_dict,  hourcol='Electricity:Facility [kW](Hourly)', hours=range(0, 24), Ftype="Hospital"):
        results = {
            "hour":[],
            "min":[],
            "max":[],
            "mu":[],
            "std":[],
        }

        for h in hours:
            results['hour'].append(h)

            results['min'].append(df_dict[h][hourcol].min())
            results['max'].append(df_dict[h][hourcol].max())

            results['mu'].append(df_dict[h][hourcol].mean())
            results['std'].append(df_dict[h][hourcol].std())

        return pd.DataFrame(results)

    
    def generate_hourly_df_dict(self, df, hourcol="hour"):
        hourly_df_dict = {}
        for f in df[hourcol]:
            hourly_df_dict[f] = df.loc[df[hourcol] == f, :]
        return hourly_df_dict


    def generate_fitted_given_hours_distros(self, H_df_dict, hourcol='Electricity:Facility [kW](Hourly)', hours=range(0, 24), Ftype="Hospital"):
        time = 0
        distGens = dict()
        for time in hours:
            distroB = H_df_dict[time][[hourcol]].dropna().values.ravel()
            dFIT = distfit()
            res = dFIT.fit_transform(distroB)
            # _, ax = dFIT.plot()
            # ax.set_title(f"Hour: {time}")
            distGens[time] = dFIT
        return distGens

    

    def generate_demand_generators(self, data_path):
        _df = pd.read_csv(data_path)
        _df["Date/Time"] = [str(s).strip() for s in _df["Date/Time"]]
        _df["Date/Time"] = self.adjust_date_col(_df, datecol="Date/Time", )
        _df["Date/Time"] = pd.to_datetime(_df["Date/Time"], format="%m/%d/%y %H:%M:%S") 
        _df["hour"] = _df["Date/Time"].dt.hour

        # store each hours data keyed on the hour (0-23)
        hourly_df_dict = self.generate_hourly_df_dict(_df, hourcol="hour")

        #  Take a look at the stats for each hour
        demand_stats = self.generate_hourly_demand_stats(hourly_df_dict)
        distro_gens = self.generate_fitted_given_hours_distros(hourly_df_dict)


        return hourly_df_dict, demand_stats, distro_gens 


# # create generic one to call 
DemSim = joblib.load("../Demand_Generators/generators/5_NodeDemandGenerator2.sav")
DemSim2 = joblib.load("../Demand_Generators/generators/7_NodeDemandGenerator.sav")

# PV130_gen2 = joblib.load("./PV_130MW_generator.sav")
PV130_GEN = joblib.load("../PV_Generators/PV_Gen/PV_130MW_generator.sav")
PV65_GEN = joblib.load("../PV_Generators/PV_Gen/PV_65MW_generator.sav")
PV38_GEN = joblib.load("../PV_Generators/PV_Gen/PV_38MW_generator.sav")
PV27_GEN = joblib.load("../PV_Generators/PV_Gen/PV_27MW_generator.sav")
PV13_GEN = joblib.load("../PV_Generators/PV_Gen/PV_13MW_generator.sav")

# Hospital_path = curdir + "/Resources/data/USA_TN_Knoxville-McGhee.Tyson.AP.723260_TMY3/RefBldgHospitalNew2004_v1.3_7.1_4A_USA_MD_BALTIMORE.csv"
# Police_path = curdir + "/Resources/data/USA_TN_Knoxville-McGhee.Tyson.AP.723260_TMY3/RefBldgLargeOfficeNew2004_v1.3_7.1_4A_USA_MD_BALTIMORE.csv"
sunlight_hours = {
        1:  [7,17],
        2:  [7,17],
        3:  [7,17],
        4:  [6,18],
        5:  [6,19],
        6:  [6,19],
        7:  [6,19],
        8:  [7,19],
        9:  [7,18],
        10: [7,18],
        11: [7,17],
        12: [7,17],
    }
sunlight_hour_monthly_ranges = [
    [[1, 2, 3, 11, 12], [8, 17]],    
    [[4], [6, 18]],    
    [[5, 6, 7], [6, 19]],    
    [[8], [7, 19]],    
    [[9, 10], [7, 18]],    
]
    
def get_sunlight_riseset(month, riseset=0, sunlight_hour_monthly_ranges=sunlight_hour_monthly_ranges):
    for group in sunlight_hour_monthly_ranges:
        months = group[0]
        hours = group[1]
        if month in months:
            return hours[riseset]
# Supermarket_path = curdir + "/Resources/data/USA_TN_Knoxville-McGhee.Tyson.AP.723260_TMY3/RefBldgSuperMarketNew2004_v1.3_7.1_4A_USA_MD_BALTIMORE.csv"
# Shelter_Path = curdir + "/Resources/data/USA_TN_Knoxville-McGhee.Tyson.AP.723260_TMY3/" + "RefBldgLargeHotelNew2004_v1.3_7.1_4A_USA_MD_BALTIMORE.csv"
# GasStation_Path = curdir + "/Resources/data/USA_TN_Knoxville-McGhee.Tyson.AP.723260_TMY3/" + "RefBldgQuickServiceRestaurantNew2004_v1.3_7.1_4A_USA_MD_BALTIMORE.csv"




# label_paths_d = {
#     "Hospital":Hospital_path,
#     "Police Station": Police_path,
#     "Supermarket":Supermarket_path,
#     "Shelter": Shelter_Path,
#     "Gas Station": GasStation_Path,
# }


# DemSim = DemandNodeSimulator(type_path_dict=label_paths_d)
# print(DemSim.get_hourly_total(1))



from setuptools import setup, find_packages
setup()
# setup(
#     name='RL_MGM',
#     version='0.0.2',
#     packages=find_packages(),
#     install_requires=[
#         # Add any dependencies here
#     ],
#     entry_points={
#         'console_scripts': [
#             'rl_mgm=rl_mgm.cli:main',
#         ],
#     },
# )



"""
IE-608 DQN example.
"""

# import
import logging
import numpy as np
from DRL import Agent, DRL_Env


def main():
    """
    mian
    """
    # --------- parameters -----------

    # number of states
    num_states = 4

    # number of plans
    num_plans = 2

    # horizon
    T = 20

    # success probability
    success_pr = [0.0, 0.5, 0.9]

    # degrade pr, implies four states
    degrade_pr = np.mat([
        [0.3, 0.5, 0.2, 0.0],
        [0.0, 0.4, 0.4, 0.2],
        [0.0, 0.0, 0.4, 0.6],
        [0.0, 0.0, 0.0, 1.0]
    ])

    # revenue
    C_r = 10

    # operating cost
    C_o = [
        i for i in range(num_states)
    ]

    # maintenance cost
    C_m = [
        i for i in range(num_plans)
    ]

    # ------------ MDP --------------

    # initial state: t, condition
    initial_state = (0, 0)

    # number of actions
    actions = range(num_plans)

    # transition function
    def trans_func(t, s, a):
        """
        transition
        """
        # terminal
        if t >= T:
            return "Delta"
        # calculate pr
        pr = []
        for i in range(num_states):
            # smaller state
            if i < s[1]:
                pr.append(
                    success_pr[a] * degrade_pr[0, i]
                )
            else:
                pr.append(np.sum([
                    success_pr[a] * degrade_pr[0, i],
                    (1 - success_pr[a]) * degrade_pr[s[1], i]
                ]))
        # sample new state according to pr
        return (t + 1, np.random.choice(
            range(num_states), size=1, replace=False,
            p=pr
        )[0])
    
    # reward function
    def reward_func(s, a, s_new):
        """
        reward
        """
        return C_r - C_o[s[1]] - C_m[a]

    # agent
    manager = Agent(
        name="manager",
        actions=actions,
        input_size=2,
        actor_hidden_layers=[100],
        critic_hidden_layers=[100],
        output_size=len(actions),
        actor_lr=1e-5,
        critic_lr=1e-5
    )

    # define problem
    problem = DRL_Env(
        name="maintenance",
        agent=manager,
        initial_state=initial_state,
        trans_func=trans_func,
        reward_func=reward_func
    )

    # logging
    logging.basicConfig(
        filename='maintenance.log', filemode='w+',
        format='%(levelname)s - %(message)s', level=logging.INFO
    )

    # DQN
    G = problem.advantage_actor_acritic(
        episodes=5000,
        discount_factor=1.0,
        write_log=True
    )

    # plot
    manager.plot_loss(dir='figs/')

    manager.plot_G(dir='figs/')

    return


if __name__ == "__main__":
    main()



# https://docs.ray.io/en/master/ray-overview/index.html

import gym, ray
import numpy as np
from gym.spaces import Discrete, Box
from ray.rllib.algorithms import ppo

import warnings

def fxn():
    warnings.warn("deprecated", DeprecationWarning)

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    fxn()
    
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    fxn()

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    fxn()

class GRIDEnv(gym.Env):
    def __init__(self, env_config, input_size=6, 
                 **kwargs):
        self.action_space = Discrete(3)
        if input_size == 6:
            self.observation_space = Box(low=np.array([0,   0,   0,   0,   0]), 
                                    #                   t,  m,   P,   D,   C, 
                                         high=np.array([24, 13, 1e7, 1e7,  1]), 
                                         dtype=np.float64)
        else:
            self.observation_space = Box(low=np.array([0, 0, 0, 0, 0, 0, 0]), 
                                    #      t,  m,   P,  D,  C, 
                                     high=np.array([24, 13, 1e7, 1e7, 1, 3]), 
                                     dtype=np.float64)
    
    def reset(self):
        return []
    def step(self, action):
        # return <obs>, <reward: float>, <done: bool>, <info: dict>
        return    [],      0.0, False,  dict()
    
env_config = {}
GRIDEnv(env_config=env_config, input_size=6)

print("DONE")

for i in range(3):
    print(i)